import datetime
from dataclasses import dataclass
from unittest.mock import MagicMock

import pytest
from wanikani_api.client import Client as WaniKaniClient
from pytest_mock import MockerFixture

from wanikani_notifier.wanikani import get_all_subjects
from wanikani_notifier.wanikani import get_available_assignments, get_notification_message, AvailableAssignments


@dataclass
class MockedLesson:
    unlocked_at = True
    started_at = False


@dataclass
class MockedReview:
    unlocked_at = True
    started_at = True


class MockedSubject:
    def __init__(self, id: int, data_updated_at: datetime):
        self.id = id
        self.data_updated_at = data_updated_at
        self._raw = {}


@pytest.fixture
def mocked_wk_client(mocker: MockerFixture) -> MagicMock:
    return mocker.Mock(spec=WaniKaniClient)


@pytest.fixture
def mocked_os(mocker: MockerFixture) -> MagicMock:
    return mocker.patch("wanikani_notifier.wanikani.os")


@pytest.fixture
def mocked_open(mocker: MockerFixture) -> MagicMock:
    return mocker.patch("builtins.open")


@pytest.fixture
def mocked_json(mocker: MockerFixture) -> MagicMock:
    return mocker.patch("wanikani_notifier.wanikani.ujson")


@pytest.fixture
def mocked_subject(mocker: MockerFixture) -> MagicMock:
    return mocker.patch("wanikani_notifier.wanikani.Subject")


def test_get_all_subjects_folder_not_found(mocked_wk_client,
                                           mocked_os, mocked_open,
                                           mocked_json):
    mocked_os.path.exists.return_value = False
    mocked_json.load.return_value = {}
    mocked_wk_client.subjects.return_value = [
        MockedSubject(id=1, data_updated_at=datetime.datetime(year=2022, month=1, day=1)),
        MockedSubject(id=2, data_updated_at=datetime.datetime(year=2021, month=1, day=1)),
        MockedSubject(id=3, data_updated_at=datetime.datetime(year=2021, month=1, day=1)),
    ]

    get_all_subjects(mocked_wk_client)

    assert mocked_os.path.exists.call_count == 2
    mocked_os.makedirs.assert_called()
    mocked_wk_client.subjects.assert_called()
    mocked_json.load.assert_not_called()
    mocked_json.dump.assert_called()


def test_get_all_subjects_folder_found_no_subject_cached(mocked_wk_client,
                                                         mocked_os, mocked_open,
                                                         mocked_json):
    mocked_os.path.exists.return_value = True
    mocked_json.load.return_value = {}
    mocked_wk_client.subjects.return_value = [
        MockedSubject(id=1, data_updated_at=datetime.datetime(year=2022, month=1, day=1)),
        MockedSubject(id=2, data_updated_at=datetime.datetime(year=2021, month=1, day=1)),
        MockedSubject(id=3, data_updated_at=datetime.datetime(year=2021, month=1, day=1)),
    ]

    get_all_subjects(mocked_wk_client)

    assert mocked_os.path.exists.call_count == 2
    mocked_os.makedirs.assert_not_called()
    mocked_wk_client.subjects.assert_called()
    mocked_json.load.assert_called()
    mocked_json.dump.assert_called()


def test_get_all_subjects_some_subjects_cached(mocked_wk_client,
                                               mocked_os, mocked_open,
                                               mocked_json, mocked_subject):
    mocked_os.path.exists.return_value = True
    mocked_json.load.return_value = [
        {"id": 1, "data_updated_at": "2021-12-27"},
        {"id": 2, "data_updated_at": "2021-12-28"},
    ]
    mocked_subject.return_value.data_updated_at = datetime.datetime(year=2021, month=12, day=12)
    mocked_wk_client.subjects.return_value = [
        MockedSubject(id=1, data_updated_at=datetime.datetime(year=2022, month=1, day=1)),
        MockedSubject(id=2, data_updated_at=datetime.datetime(year=2021, month=1, day=1)),
        MockedSubject(id=3, data_updated_at=datetime.datetime(year=2021, month=1, day=1)),
    ]

    get_all_subjects(mocked_wk_client)

    assert mocked_os.path.exists.call_count == 2
    mocked_os.makedirs.assert_not_called()
    mocked_wk_client.subjects.assert_called()
    mocked_json.load.assert_called()
    mocked_json.dump.assert_called()


def test_available_assignments_none(mocked_wk_client):
    mocked_wk_client.assignments.return_value = []

    get_available_assignments(mocked_wk_client, {}, datetime.datetime.utcnow())

    mocked_wk_client.assignments.assert_called()


def test_get_message_no_new_reviews_no_new_lessons():
    message = get_notification_message(AvailableAssignments(reviews=0, lessons=0))

    assert message is None


def test_get_message_new_reviews_no_new_lessons():
    message = get_notification_message(AvailableAssignments(reviews=2, lessons=0))

    assert "lesson" not in message
    assert "review" in message


def test_get_message_no_new_reviews_new_lessons():
    message = get_notification_message(AvailableAssignments(reviews=0, lessons=3))

    assert "lesson" in message
    assert "review" not in message


def test_get_message_new_reviews_new_lessons():
    message = get_notification_message(AvailableAssignments(reviews=2, lessons=3))

    assert "lesson" in message
    assert "review" in message



from typing import Any, Dict
from unittest.mock import MagicMock

import pytest
from pytest_mock import MockerFixture

from tests.notifiers.test_notifier import NotifierTester
from wanikani_notifier.notifiers.console import ConsoleNotifier
from wanikani_notifier.notifiers.notifier import Notifier


@pytest.fixture
def mocked_console_print(mocker: MockerFixture) -> MagicMock:
    return mocker.patch("builtins.print")


class TestConsoleNotifier(NotifierTester):
    @property
    def key(self) -> str:
        return "console"

    @pytest.fixture
    def build_parameters(self) -> Dict[str, Any]:
        return {}

    @pytest.fixture
    def imp(self, mocked_console_print) -> Notifier:
        yield ConsoleNotifier()

    def test_notify_simple_message(self, imp, mocked_console_print):
        super().test_notify_simple_message(imp)
        mocked_console_print.assert_called()

    def test_notify_message_all_options(self, imp, mocked_console_print):
        super().test_notify_message_all_options(imp)
        mocked_console_print.assert_called()



from abc import ABC, abstractmethod
from typing import Callable, Optional


class NoMessageProvided(RuntimeError):
    pass


class Notifier(ABC):
    @property
    @abstractmethod
    def key(self) -> str:
        """
        Defines the key used to identify a particular notifier class.
        :return: the key string identifying a particular notifier class.
        """

    @abstractmethod
    def notify(self, title: str, message: str, url: Optional[str] = None, icon: Optional[str] = None):
        """
        Sends a notification.

        :param title: Title of the notification to send.
        :param message: Content of the notification to send.
        :param url: Optional url to display.
        :param icon: Optional icon to display
        """


class NotifierFactory:
    def __init__(self):
        self._builders = {}

    def register_builder(self, key: str, builder: Callable) -> None:
        self._builders[key] = builder

    def create(self, key: str, **kwargs) -> Notifier:
        return self._builders[key.lower()](**kwargs)


factory = NotifierFactory()



from typing import Optional

import pushsafer

from wanikani_notifier.notifiers.notifier import Notifier, NoMessageProvided, factory


class PushSaferNotifier(Notifier):
    def __init__(self, private_key: str):
        self._client = pushsafer.Client(private_key)

    @classmethod
    def key(cls) -> str:
        return "pushsafer"

    @classmethod
    def build(cls, private_key: str, **_ignored) -> "PushSaferNotifier":
        return PushSaferNotifier(private_key)

    def notify(self, title: str, message: str, url: Optional[str] = None, icon: Optional[str] = None):
        if not message:
            raise NoMessageProvided

        self._client.send_message(title=title, message=message, url=url)


factory.register_builder(PushSaferNotifier.key(), PushSaferNotifier.build)



import os
from datetime import datetime
from collections import namedtuple
from typing import Optional, Dict

import pytz
import ujson as ujson
from wanikani_api.client import Client as WaniKaniClient
from wanikani_api.models import Subject

AvailableAssignments = namedtuple("AvailableAssignments", ("reviews", "lessons"))


SUBJECTS_CACHED_FILENAME = "subjects.json"


def get_all_subjects(wk_client: WaniKaniClient) -> Dict[int, Subject]:
    all_subjects: Dict[int, Subject] = {}

    if not os.path.exists("data/"):
        os.makedirs("data/")

    if os.path.exists(f"data/{SUBJECTS_CACHED_FILENAME}"):
        subject_jsons = ujson.load(open(f"data/{SUBJECTS_CACHED_FILENAME}", "r"))
        all_subjects = {subject_json["id"]: Subject(subject_json) for subject_json in subject_jsons}

    latest_update = max([s.data_updated_at for s in all_subjects.values()]) if all_subjects else datetime.min
    for subject in wk_client.subjects(updated_after=latest_update.strftime("%Y-%m-%dT%H:%M:%S.%f"), fetch_all=True):
        all_subjects[subject.id] = subject

    ujson.dump([subject._raw for subject in all_subjects.values()], open(f"data/{SUBJECTS_CACHED_FILENAME}", "w"))

    return all_subjects


def get_available_assignments(wanikani_client: WaniKaniClient,
                              all_subjects: Dict[int, Subject],
                              end: datetime,
                              start: datetime = None
                              ) -> AvailableAssignments:
    """
    Gets the number of reviews and lessons that are available in the provided time period.

    :rtype: object
    :param wanikani_client: WaniKani client to use for fetching assignments
    :param start: Start of the time period when assignments are considered (inclusive).
    :param end: End of the time period when assignments are considered (inclusive).
    :return: the available assignments.
    """
    user_level = wanikani_client.user_information().level
    review_count = sum(1
                       for a in wanikani_client.assignments(fetch_all=True, unlocked=True, started=True,
                                                            available_after=start, available_before=end)
                       if not a.hidden
                       and a.subject_id in all_subjects and all_subjects[a.subject_id].level <= user_level
                       )
    lesson_count = sum(1
                       for a in wanikani_client.assignments(fetch_all=True, unlocked=True, started=False)
                       if a.created_at.replace(tzinfo=pytz.utc) <= end.replace(tzinfo=pytz.utc)
                       and (start.replace(tzinfo=pytz.utc) <= a.created_at.replace(tzinfo=pytz.utc) if start else True)
                       )

    return AvailableAssignments(reviews=review_count, lessons=lesson_count)


def get_notification_message(available_assignments: AvailableAssignments,
                             message_template: Optional[str] = None) -> Optional[str]:
    """
    Creates the notification message corresponding to the available assignments.

    :param available_assignments: Number of reviews and lessons that are available.
    :param message_template: Message template where the generated number of reviews and lessons
                                replaces the {} format specifier
    :return: a string containing the notification message to send if at least one assignment is available,
                None otherwise
    """
    if not sum(available_assignments):
        return

    available_lessons_message = ""
    available_reviews_message = ""
    if available_assignments.lessons:
        available_lessons_message = f"{available_assignments.lessons} lessons"
    if available_assignments.reviews:
        available_reviews_message = f"{available_assignments.reviews} reviews"

    full_message = " and ".join(m for m in [available_lessons_message, available_reviews_message] if m)
    return message_template.format(full_message) if message_template else full_message



from typing import Optional

from wanikani_notifier.notifiers.notifier import Notifier, NoMessageProvided, factory


class ConsoleNotifier(Notifier):
    @classmethod
    def build(cls, **_ignored) -> "ConsoleNotifier":
        return ConsoleNotifier()

    @classmethod
    def key(cls) -> str:
        return "console"

    def notify(self, title: str, message: str, url: Optional[str] = None, icon: Optional[str] = None):
        if not message:
            raise NoMessageProvided

        notification = f"{title}:\n{message}"
        if url:
            notification += f"\n{url}"

        print(notification)


factory.register_builder(ConsoleNotifier.key(), ConsoleNotifier.build)
